/* ------------------------------------------------------------------------ */
/* hmnull.c - Funktionen zur numerischen Bestimmung von Nullstellen         */
/*                                                                          */
/*                  Copyright Tilman Kuepper 2016.                          */
/*    Distributed under the Boost Software License, Version 1.0.            */
/*        (See accompanying file LICENSE_1_0.txt or copy at                 */
/*               http://www.boost.org/LICENSE_1_0.txt)                      */
/*                                                                          */
/* Letzte Bearbeitung: 12. Mai 2016                                         */
/* ------------------------------------------------------------------------ */

#include "hmnull.h"
#include <math.h>
#include <assert.h>

/* ------------------------------------------------------------------------ */
/* Rueckgabewerte bzw. Fehlermeldungen in Textform.                         */
/* ------------------------------------------------------------------------ */
const char *hmnull_errmsg[] =
{
    "OK",                       /* HMNULL_OK         */
    "Iteration failed",         /* HMNULL_ERR_ITFAIL */
    "Invalid parameter",        /* HMNULL_ERR_PARAM  */
    "Overflow error"            /* HMNULL_ERR_NAN    */
};

/* ------------------------------------------------------------------------ */
/* Interne Hilfsfunktion: Vorzeichen einer double-Zahl bestimmen            */
/* ------------------------------------------------------------------------ */
static int sgn_(double x)
{
    if(x < 0) return -1;
    if(x > 0) return +1;
    return 0;
}

/* ------------------------------------------------------------------------ */
/* Interne Hilfsfunktion: Ueberlauf (NaN) eines double-Werts feststellen    */
/* ------------------------------------------------------------------------ */
static int is_nan_(double x)
{
    /*  "A comparison with a NaN always returns an unordered result even
     *  when comparing with itself. The comparison predicates are either
     *  signaling or non-signaling; the signaling versions signal the
     *  invalid operation exception for such comparisons. The equality
     *  and inequality predicates are non-signaling so x = x returning
     *  false can be used to test if x is a quiet NaN."
     *
     *  https://en.wikipedia.org/w/index.php?title=NaN&oldid=704603948 
     *  last visited Mar. 11, 2016.
     */

    return (x != x) ? 1 : 0;
}

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Bisektionsverfahren: Als  */
/* Parameter "fn" ist die Funktion zu uebergeben, deren Nullstelle bestimmt */
/* werden soll (vergl. "typedef hmnull_fn" in "hmnull.h"). Die Nullstelle   */
/* muss zwischen x1...x2 liegen, fn(x1) und fn(x2) muessen unterschiedliche */
/* Vorzeichen haben. Das Verfahren wird beendet, falls eine Loesung mit der */
/* Genauigkeit "epsilon" ermittelt oder nach "max_iter" Iterationen keine   */
/* Loesung gefunden wurde. Ueber "pdata" kann ein beliebiger Zeiger an die  */
/* Implementierung von "fn" weitergeleitet werden. Ueber "errorcode" wird   */
/* ein Fehlercode zurueckgegeben, moegliche Codes sind: HMNULL_ERR_ITFAIL,  */
/* HMNULL_ERR_PARAM, HMNULL_ERR_NAN und HMNULL_OK. (Es ist zulaessig, im    */
/* Parameter "errcode" NULL zu uebergeben, falls kein Fehlercode benoetigt  */
/* wird.) Die berechnete Nullstelle wird als Rueckgabewert zurueckgegeben.  */
/* ------------------------------------------------------------------------ */
double bisect_ex(hmnull_fn fn, double x1, double x2,
    double epsilon, size_t max_iter, int *errorcode, void *pdata)
{
    size_t iter = 0;
    double fn_x1 = fn(x1, pdata), fn_x2 = fn(x2, pdata);

    assert(max_iter > 0);
    if(errorcode) *errorcode = HMNULL_OK;

    /* Ueberlauf schon bei den Startwerten -> Abbruch! */
    if(is_nan_(fn_x1) || is_nan_(fn_x2))
    { if(errorcode) *errorcode = HMNULL_ERR_NAN; return 0; }

    /* Sonderfall: Nullstelle liegt auf der Intervallgrenze x1 oder x2 */
    if(fabs(fn_x1) <= fabs(epsilon) || fabs(fn_x2) <= fabs(epsilon))
        return fabs(fn_x1) < fabs(fn_x2) ? x1 : x2;

    /* Fehler: fn(x1) und fn(x2) haben gleiches Vorzeichen -> Abbruch! */
    if(sgn_(fn_x1) * sgn_(fn_x2) >= 0)
    { if(errorcode) *errorcode = HMNULL_ERR_PARAM; return 0; }

    /* Hier findet das eigentliche Bisektionsverfahren statt */
    for(iter = 0; iter < max_iter; ++iter)
    {
        double x = 0.5 * x1 + 0.5 * x2, fn_x = fn(x, pdata);
        if(is_nan_(x) || is_nan_(fn_x))  /* Ueberlauf/NaN? */
        { if(errorcode) *errorcode = HMNULL_ERR_NAN; return 0; }

        if(fabs(fn_x) <= fabs(epsilon)) return x;  /* Treffer! */

        /* Naechster Iterationsschritt */
        if(sgn_(fn_x) * sgn_(fn_x1) < 0)
            x2 = x, fn_x2 = fn_x;
        else
            x1 = x, fn_x1 = fn_x;
    }

    /* Fehler: Maximale Iterationsanzahl wurde erreicht */
    if(errorcode) *errorcode = HMNULL_ERR_ITFAIL;
    return 0;
}

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Bisektionsverfahren:      */
/* Siehe Dokumentation von "bisect_ex" fuer Details zum Funktionsaufruf.    */
/* Im Gegensatz zu "bisect_ex" werden hier fuer die Parameter "epsilon",    */
/* "max_iter", "errcode" und "pdata" folgende Standardwerte uebergeben:     */
/* HMNULL_EPSILON, HMNULL_MAXITER, NULL und NULL.                           */
/* ------------------------------------------------------------------------ */
double bisect(hmnull_fn fn, double x1, double x2)
{
    double x = 0;
    int err = HMNULL_OK;

    x = bisect_ex(fn, x1, x2, HMNULL_EPSILON, HMNULL_MAXITER, &err, NULL);
    if(err == HMNULL_ERR_ITFAIL) assert(!"bisect: iteration failed");
    if(err == HMNULL_ERR_PARAM)  assert(!"bisect: f(x1), f(x2) same sign");
    if(err == HMNULL_ERR_NAN)    assert(!"bisect: overflow error");
    return x;
}

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Newtonverfahren: Als      */
/* Parameter "fn" ist die Funktion zu uebergeben, deren Nullstelle bestimmt */
/* werden soll (vergl. "typedef hmnull_fn" in "hmnull.h"). In "deriv" ist   */
/* die Ableitung von "fn" zu uebergeben. Der Algorithmus beginnt mit dem    */
/* Startwert "x0". Er wird beendet, falls eine Loesung mit der Genauigkeit  */
/* "epsilon" ermittelt oder nach "max_iter" Iterationen noch keine Loesung  */
/* gefunden wurde. Ueber "pdata" kann ein Zeiger an die Funktionen "fn" und */
/* "deriv" weitergeleitet werden. Ueber "errorcode" kann ein Fehlercode an  */
/* den Aufrufer zurueckgegeben werden, moegliche Codes: HMNULL_ERR_ITFAIL,  */
/* HMNULL_ERR_NAN und HMNULL_OK. (Es ist zulaessig, im Parameter "errcode"  */
/* 0 zu uebergeben, falls kein Fehlercode benoetigt wird.) Die berechnete   */
/* Nullstelle wird als Rueckgabewert zurueckgegeben.                        */
/* ------------------------------------------------------------------------ */
double newton_ex(hmnull_fn fn, hmnull_fn deriv, double x0,
    double epsilon, size_t max_iter, int *errorcode, void *pdata)
{
    size_t iter = 0;
    double fn_x0 = fn(x0, pdata);

    assert(max_iter > 0);
    if(errorcode) *errorcode = HMNULL_OK;

    /* Ueberlauf schon beim Startwert -> Abbruch! */
    if(is_nan_(fn_x0))
    { if(errorcode) *errorcode = HMNULL_ERR_NAN; return 0; }

    /* Sonderfall: Nullstelle liegt direkt bei x0 */
    if(fabs(fn_x0) <= fabs(epsilon)) return x0;

    /* Hier findet das eigentliche Newtonverfahren statt */
    for(iter = 0; iter < max_iter; ++iter)
    {
        double x1 = x0, deriv_x0 = deriv(x0, pdata);
        if(deriv_x0 == 0 || is_nan_(deriv_x0))  /* Ueberlauf/NaN? */
        { if(errorcode) *errorcode = HMNULL_ERR_NAN; return 0; }

        x1 -= fn_x0 / deriv_x0;
        if(is_nan_(x1))  /* Ueberlauf bei Division -> Abbruch! */
        { if(errorcode) *errorcode = HMNULL_ERR_NAN; return 0; }

        x0 = x1, fn_x0 = fn(x0, pdata);
        if(is_nan_(fn_x0))  /* Ueberlauf bei Funktionswert-Berechnung? */
        { if(errorcode) *errorcode = HMNULL_ERR_NAN; return 0; }

        if(fabs(fn_x0) <= fabs(epsilon)) return x0;  /* Treffer! */
    }

    /* Fehler: Maximale Iterationsanzahl wurde erreicht */
    if(errorcode) *errorcode = HMNULL_ERR_ITFAIL;
    return 0;
}

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Newtonverfahren:          */
/* Siehe Dokumentation von "newton_ex" fuer Details zum Funktionsaufruf.    */
/* Im Gegensatz zu "newton_ex" werden hier fuer die Parameter "epsilon",    */
/* "max_iter", "errcode" und "pdata" folgende Standardwerte uebergeben:     */
/* HMNULL_EPSILON, HMNULL_MAXITER, NULL und NULL.                           */
/* ------------------------------------------------------------------------ */
double newton(hmnull_fn fn, hmnull_fn deriv, double x0)
{
    double x = 0;
    int err = HMNULL_OK;

    x = newton_ex(fn, deriv, x0, HMNULL_EPSILON, HMNULL_MAXITER, &err, NULL);
    if(err == HMNULL_ERR_ITFAIL) assert(!"newton: iteration failed");
    if(err == HMNULL_ERR_NAN)    assert(!"newton: overflow error");
    return x;
}

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Sekantenverfahren: Als    */
/* Parameter "fn" ist die Funktion zu uebergeben, deren Nullstelle bestimmt */
/* werden soll (vergl. "typedef hmnull_fn" in "hmnull.h"). Die Nullstelle   */
/* sollte dabei moeglichst in der Naehe des Intervalls x1...x2 liegen. Das  */
/* Verfahren wird beendet, falls eine Loesung mit der Genauigkeit "epsilon" */
/* ermittelt oder nach "max_iter" Iterationen keine Loesung gefunden wurde. */
/* Ueber "pdata" kann ein beliebiger Zeiger an "fn" weitergeleitet werden.  */
/* Ueber den Zeiger "errorcode" kann ein Fehlercode zurueckgegeben werden,  */
/* moegliche Codes sind: HMNULL_ERR_ITFAIL, HMNULL_ERR_NAN und HMNULL_OK.   */
/* (Als "errcode" kann 0 uebergeben werden, falls kein Fehlercode benoetigt */
/* wird.) Die berechnete Nullstelle wird als Rueckgabewert zurueckgegeben.  */
/* ------------------------------------------------------------------------ */
double secant_ex(hmnull_fn fn, double x1, double x2,
    double epsilon, size_t max_iter, int *errorcode, void *pdata)
{
    size_t iter = 0;
    double fn_x1 = fn(x1, pdata), fn_x2 = fn(x2, pdata);

    assert(max_iter > 0);
    if(errorcode) *errorcode = HMNULL_OK;

    /* Ueberlauf schon bei den Startwerten -> Abbruch! */
    if(is_nan_(fn_x1) || is_nan_(fn_x2))
    { if(errorcode) *errorcode = HMNULL_ERR_NAN; return 0; }

    /* Sonderfall: Nullstelle liegt auf der Intervallgrenze x1 oder x2 */
    if(fabs(fn_x1) <= fabs(epsilon) || fabs(fn_x2) <= fabs(epsilon))
        return fabs(fn_x1) < fabs(fn_x2) ? x1 : x2;

    /* Hier findet das eigentliche Sekantenverfahren statt */
    for(iter = 0; iter < max_iter; ++iter)
    {
        double x3 = x2 - fn_x2 * (x2 - x1) / (fn_x2 - fn_x1);
        double fn_x3 = fn(x3, pdata);

        if(is_nan_(x3) || is_nan_(fn_x3))  /* Ueberlauf/NaN? */
        { if(errorcode) *errorcode = HMNULL_ERR_NAN; return 0; }

        if(fabs(fn_x3) <= fabs(epsilon)) return x3;  /* Treffer! */

        /* Naechster Iterationsschritt */
        x1 = x2, fn_x1 = fn_x2;
        x2 = x3, fn_x2 = fn_x3;
    }

    /* Fehler: Maximale Iterationsanzahl wurde erreicht */
    if(errorcode) *errorcode = HMNULL_ERR_ITFAIL;
    return 0;
}

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Sekantenverfahren:        */
/* Siehe Dokumentation von "secant_ex" fuer Details zum Funktionsaufruf.    */
/* Im Gegensatz zu "secant_ex" werden hier fuer die Parameter "epsilon",    */
/* "max_iter", "errcode" und "pdata" folgende Standardwerte uebergeben:     */
/* HMNULL_EPSILON, HMNULL_MAXITER, NULL und NULL.                           */
/* ------------------------------------------------------------------------ */
double secant(hmnull_fn fn, double x1, double x2)
{
    double x = 0;
    int err = HMNULL_OK;

    x = secant_ex(fn, x1, x2, HMNULL_EPSILON, HMNULL_MAXITER, &err, NULL);
    if(err == HMNULL_ERR_ITFAIL) assert(!"secant: iteration failed");
    if(err == HMNULL_ERR_NAN)    assert(!"newton: overflow error");
    return x;
}
