/* ------------------------------------------------------------------------ */
/* hmnull.h - Funktionen zur numerischen Bestimmung von Nullstellen         */
/*                                                                          */
/*                  Copyright Tilman Kuepper 2016.                          */
/*    Distributed under the Boost Software License, Version 1.0.            */
/*        (See accompanying file LICENSE_1_0.txt or copy at                 */
/*               http://www.boost.org/LICENSE_1_0.txt)                      */
/*                                                                          */
/* Letzte Bearbeitung: 12. Mai 2016                                         */
/* ------------------------------------------------------------------------ */

#ifndef HMNULL_H_Included
#define HMNULL_H_Included

#include <stddef.h>  /* size_t */

#ifdef __cplusplus
extern "C" {
#endif

/* ------------------------------------------------------------------------ */
/* Rueckgabewerte bzw. Fehlermeldungen von Funktionen                       */
/* ------------------------------------------------------------------------ */
#define HMNULL_OK         0    /* Funktion wurde erfolgreich beendet        */
#define HMNULL_ERR_ITFAIL 1    /* Berechnungsverfahren konvergiert nicht    */
#define HMNULL_ERR_PARAM  2    /* Ungueltige Parameter beim Funktionsaufruf */
#define HMNULL_ERR_NAN    3    /* Ueberlauf/NaN waehrend der Berechnung     */

/* Rueckgabewerte bzw. Fehlermeldungen in Textform. */
extern const char *hmnull_errmsg[];

/* ------------------------------------------------------------------------ */
/* Gewuenschte Genauigkeit                                                  */
/* ------------------------------------------------------------------------ */
#ifndef HMNULL_EPSILON
#define HMNULL_EPSILON 1.0e-10
#endif

/* ------------------------------------------------------------------------ */
/* Maximale Anzahl von Iterationen des Loesungsverfahrens                   */
/* ------------------------------------------------------------------------ */
#ifndef HMNULL_MAXITER
#define HMNULL_MAXITER 10000UL
#endif

/* ------------------------------------------------------------------------ */
/* Funktionen mit einer Veraenderlichen (x-Wert), die an HMNull uebergeben  */
/* werden sollen, muessen in der folgenden Form implementiert werden. Der   */
/* letzte Parameter (pdata) dient dazu, auf Wunsch zusaetzlich zum x-Wert   */
/* weitere Informationen an die Implementierung der Funktion zu uebergeben. */
/* ------------------------------------------------------------------------ */
typedef double hmnull_fn(double x, void *pdata);

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Bisektionsverfahren: Als  */
/* Parameter "fn" ist die Funktion zu uebergeben, deren Nullstelle bestimmt */
/* werden soll (vergl. "typedef hmnull_fn" in "hmnull.h"). Die Nullstelle   */
/* muss zwischen x1...x2 liegen, fn(x1) und fn(x2) muessen unterschiedliche */
/* Vorzeichen haben. Das Verfahren wird beendet, falls eine Loesung mit der */
/* Genauigkeit "epsilon" ermittelt oder nach "max_iter" Iterationen keine   */
/* Loesung gefunden wurde. Ueber "pdata" kann ein beliebiger Zeiger an die  */
/* Implementierung von "fn" weitergeleitet werden. Ueber "errorcode" wird   */
/* ein Fehlercode zurueckgegeben, moegliche Codes sind: HMNULL_ERR_ITFAIL,  */
/* HMNULL_ERR_PARAM, HMNULL_ERR_NAN und HMNULL_OK. (Es ist zulaessig, im    */
/* Parameter "errcode" NULL zu uebergeben, falls kein Fehlercode benoetigt  */
/* wird.) Die berechnete Nullstelle wird als Rueckgabewert zurueckgegeben.  */
/* ------------------------------------------------------------------------ */
double bisect_ex(hmnull_fn fn, double x1, double x2,
    double epsilon, size_t max_iter, int *errorcode, void *pdata);

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Bisektionsverfahren:      */
/* Siehe Dokumentation von "bisect_ex" fuer Details zum Funktionsaufruf.    */
/* Im Gegensatz zu "bisect_ex" werden hier fuer die Parameter "epsilon",    */
/* "max_iter", "errcode" und "pdata" folgende Standardwerte uebergeben:     */
/* HMNULL_EPSILON, HMNULL_MAXITER, NULL und NULL.                           */
/* ------------------------------------------------------------------------ */
double bisect(hmnull_fn fn, double x1, double x2);

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Newtonverfahren: Als      */
/* Parameter "fn" ist die Funktion zu uebergeben, deren Nullstelle bestimmt */
/* werden soll (vergl. "typedef hmnull_fn" in "hmnull.h"). In "deriv" ist   */
/* die Ableitung von "fn" zu uebergeben. Der Algorithmus beginnt mit dem    */
/* Startwert "x0". Er wird beendet, falls eine Loesung mit der Genauigkeit  */
/* "epsilon" ermittelt oder nach "max_iter" Iterationen noch keine Loesung  */
/* gefunden wurde. Ueber "pdata" kann ein Zeiger an die Funktionen "fn" und */
/* "deriv" weitergeleitet werden. Ueber "errorcode" kann ein Fehlercode an  */
/* den Aufrufer zurueckgegeben werden, moegliche Codes: HMNULL_ERR_ITFAIL,  */
/* HMNULL_ERR_NAN und HMNULL_OK. (Es ist zulaessig, im Parameter "errcode"  */
/* 0 zu uebergeben, falls kein Fehlercode benoetigt wird.) Die berechnete   */
/* Nullstelle wird als Rueckgabewert zurueckgegeben.                        */
/* ------------------------------------------------------------------------ */
double newton_ex(hmnull_fn fn, hmnull_fn deriv, double x0,
    double epsilon, size_t max_iter, int *errorcode, void *pdata);

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Newtonverfahren:          */
/* Siehe Dokumentation von "newton_ex" fuer Details zum Funktionsaufruf.    */
/* Im Gegensatz zu "newton_ex" werden hier fuer die Parameter "epsilon",    */
/* "max_iter", "errcode" und "pdata" folgende Standardwerte uebergeben:     */
/* HMNULL_EPSILON, HMNULL_MAXITER, NULL und NULL.                           */
/* ------------------------------------------------------------------------ */
double newton(hmnull_fn fn, hmnull_fn deriv, double x0);

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Sekantenverfahren: Als    */
/* Parameter "fn" ist die Funktion zu uebergeben, deren Nullstelle bestimmt */
/* werden soll (vergl. "typedef hmnull_fn" in "hmnull.h"). Die Nullstelle   */
/* sollte dabei moeglichst in der Naehe des Intervalls x1...x2 liegen. Das  */
/* Verfahren wird beendet, falls eine Loesung mit der Genauigkeit "epsilon" */
/* ermittelt oder nach "max_iter" Iterationen keine Loesung gefunden wurde. */
/* Ueber "pdata" kann ein beliebiger Zeiger an "fn" weitergeleitet werden.  */
/* Ueber den Zeiger "errorcode" kann ein Fehlercode zurueckgegeben werden,  */
/* moegliche Codes sind: HMNULL_ERR_ITFAIL, HMNULL_ERR_NAN und HMNULL_OK.   */
/* (Als "errcode" kann 0 uebergeben werden, falls kein Fehlercode benoetigt */
/* wird.) Die berechnete Nullstelle wird als Rueckgabewert zurueckgegeben.  */
/* ------------------------------------------------------------------------ */
double secant_ex(hmnull_fn fn, double x1, double x2,
    double epsilon, size_t max_iter, int *errorcode, void *pdata);

/* ------------------------------------------------------------------------ */
/* Numerische Bestimmung einer Nullstelle mittels Sekantenverfahren:        */
/* Siehe Dokumentation von "secant_ex" fuer Details zum Funktionsaufruf.    */
/* Im Gegensatz zu "secant_ex" werden hier fuer die Parameter "epsilon",    */
/* "max_iter", "errcode" und "pdata" folgende Standardwerte uebergeben:     */
/* HMNULL_EPSILON, HMNULL_MAXITER, NULL und NULL.                           */
/* ------------------------------------------------------------------------ */
double secant(hmnull_fn fn, double x1, double x2);

#ifdef __cplusplus
}
#endif

/* ------------------------------------------------------------------------ */
/* In C++-Projekten stehen die zusaetzlichen Funktions-Templates Bisect(),  */
/* Secant() und Newton() zur Verfuegung, wodurch die Nullstellenbestimmung  */
/* im Vergleich zur Programmierung in C noch weiter vereinfacht wird.       */
/* ------------------------------------------------------------------------ */
#if defined(__cplusplus) && !defined(HMNULL_NO_CPP)

#include <stdexcept>  /* std::runtime_error */

/* Namespace fuer Nullstellenberechnung ("HM" = "Hochschule Muenchen") */
namespace HMNull
{
    /* Interne Hilfsfunktionen */
    namespace Internal
    {
        class DummyT { };

        /*  Die vom Anwender uebergebenen Callbacks zur Berechnung des
         *  Funktionswerts und der ersten Ableitung werden ueber die
         *  Struktur "CallbackList" an FunctProxy() und DerivProxy()
         *  weitergeleitet und dann von dort aus aufgerufen.
         */
        template<typename FunctT, typename DerivT>
        struct CallbackList
        {
            FunctT *pFunct;  /* CB zur Berechnung des Funktionswerts   */
            DerivT *pDeriv;  /* CB zur Berechnung der ersten Ableitung */
        };

        /*  Aus der in "pdata" uebergebenen "CallbackList" wird der erste
         *  Callback (zur Berechnung von Funktionswerten) aufgerufen und
         *  der berechnete Funktionswert als Rueckgabewert zurueckgegeben.
         */
        template<typename FunctT, typename DerivT>
        static double FunctProxy(double x, void *pdata)
        {
            typedef CallbackList<FunctT, DerivT> CallbackListT;
            CallbackListT *pCallbackList = (CallbackListT*)pdata;

            /* assert(pCallbackList && pCallbackList->pFunct); */
            return (*pCallbackList->pFunct)(x);
        }

        /*  Aus der in "pdata" uebergebenen "CallbackList" wird der zweite
         *  Callback (zur Berechnung der ersten Ableitung) aufgerufen und
         *  die berechnete Ableitung als Rueckgabewert zurueckgegeben.
         */
        template<typename FunctT, typename DerivT>
        static double DerivProxy(double x, void *pdata)
        {
            typedef CallbackList<FunctT, DerivT> CallbackListT;
            CallbackListT *pCallbackList = (CallbackListT*)pdata;

            /* assert(pCallbackList && pCallbackList->pDeriv); */
            return (*pCallbackList->pDeriv)(x);
        }
    }

    /* -------------------------------------------------------------------- */
    /* Nullstellenbestimmung mit dem Bisektionsverfahren: Die Funktion f(x) */
    /* ist als Funktionsobjekt, Lambda-Ausdruck oder als Funktionszeiger    */
    /* zu uebergeben. Zusaetzlich ist ein Intervall x1...x2 anzugeben, wo   */
    /* die zu suchende Nullstelle (oder eine groessere ungerade Anzahl von  */
    /* Nullst.) liegen muss. Optional koennen die erforderliche Genauigkeit */
    /* sowie die maximal zulaessige Iterationsanzahl eingestellt werden.    */
    /* -------------------------------------------------------------------- */
    template<typename FunctT> double Bisect(
        FunctT funct, double x1, double x2,
        double epsilon = HMNULL_EPSILON, size_t maxiter = HMNULL_MAXITER)
    {
        using namespace Internal;
        CallbackList<FunctT, DummyT> callbackList;
        callbackList.pFunct = &funct, callbackList.pDeriv = NULL;

        int err = HMNULL_OK;
        double x0 = bisect_ex(
            FunctProxy<FunctT, DummyT>,  /* Proxy zur f(x)-Berechnung */
            x1, x2, epsilon, maxiter, &err, &callbackList);

        if(err != HMNULL_OK) throw std::runtime_error(hmnull_errmsg[err]);
        return x0;
    }

    /* -------------------------------------------------------------------- */
    /* Nullstellenbestimmung mit dem Newtonverfahren: Die Funktion f(x) und */
    /* deren Ableitung f'(x) sind als Funktionsobjekte, Lambda-Ausdruecke   */
    /* oder als Funktionszeiger zu uebergeben. Weiterhin ist der Startwert  */
    /* x0 fuer das Iterationsverfahren anzugeben. Optional koennen die zu   */
    /* erreichende Genauigkeit sowie die max. zulaessige Iterationsanzahl   */
    /* eingestellt werden.                                                  */
    /* -------------------------------------------------------------------- */
    template<typename FunctT, typename DerivT> double Newton(
        FunctT funct, DerivT deriv, double x0,
        double epsilon = HMNULL_EPSILON, size_t maxiter = HMNULL_MAXITER)
    {
        using namespace Internal;
        CallbackList<FunctT, DerivT> callbackList;
        callbackList.pFunct = &funct, callbackList.pDeriv = &deriv;

        int err = HMNULL_OK;
        x0 = newton_ex(
            FunctProxy<FunctT, DerivT>,  /* Proxy zur Berechnung von f(x)  */
            DerivProxy<FunctT, DerivT>,  /* Proxy zur Berechnung von f'(x) */
            x0, epsilon, maxiter, &err, &callbackList);

        if(err != HMNULL_OK) throw std::runtime_error(hmnull_errmsg[err]);
        return x0;
    }

    /* -------------------------------------------------------------------- */
    /* Nullstellenbestimmung mit dem Sekantenverfahren: Die Funktion f(x)   */
    /* ist als Funktionsobjekt, Lambda-Ausdruck oder als Funktionszeiger    */
    /* zu uebergeben. Zusaetzlich ist ein Intervall x1...x2 anzugeben, in   */
    /* dessen Naehe die gesuchte Nullstelle liegen sollte. Optional koennen */
    /* die gewuenschte Genauigkeit sowie die maximal zulaessige Anzahl von  */
    /* Iterationen eingestellt werden.                                      */
    /* -------------------------------------------------------------------- */
    template<typename FunctT> double Secant(
        FunctT funct, double x1, double x2,
        double epsilon = HMNULL_EPSILON, size_t maxiter = HMNULL_MAXITER)
    {
        using namespace Internal;
        CallbackList<FunctT, DummyT> callbackList;
        callbackList.pFunct = &funct, callbackList.pDeriv = NULL;

        int err = HMNULL_OK;
        double x0 = secant_ex(
            FunctProxy<FunctT, DummyT>,  /* Proxy zur Berechnung von f(x) */
            x1, x2, epsilon, maxiter, &err, &callbackList);

        if(err != HMNULL_OK) throw std::runtime_error(hmnull_errmsg[err]);
        return x0;
    }
}  /* namespace HMNull */

#endif  /* __cplusplus       */
#endif  /* HMNULL_H_Included */
